#!/usr/bin/env python
# Copyright (c) 2006 ActiveState Software Inc.
# See LICENSE.txt for license details.

"""Completion evaluation code for JavaScript"""

import types
import re
from pprint import pformat

from codeintel2.common import *
from codeintel2.util import indent
from codeintel2.tree import TreeEvaluator


class CandidatesForTreeEvaluator(TreeEvaluator):
    def _elem_from_scoperef(self, scoperef):
        """A scoperef is (<blob>, <lpath>). Return the actual elem in
        the <blob> ciElementTree being referred to.
        """
        elem = scoperef[0]
        for lname in scoperef[1]:
            elem = elem.names[lname]
        return elem

    def _tokenize_citdl_expr(self, expr):
        for tok in expr.split('.'):
            if tok.endswith('()'):
                yield tok[:-2]
                yield '()'
            else:
                yield tok
    def _join_citdl_expr(self, tokens):
        return '.'.join(tokens).replace('.()', '()')

class JavaScriptTreeEvaluator(CandidatesForTreeEvaluator):
    def eval_cplns(self):
        self.log_start()
        start_scoperef = self.get_start_scoperef()
        self.info("start scope is %r", start_scoperef)
        hits = self._hits_from_citdl(self.expr, start_scoperef)
        # For logging messages every call
        #msg = '\n'.join((m % args) for lvl,m,args in self.ctlr.log)
        #print indent(msg)
        #print indent('\n'.join(["Hit: %r" % (hit, ) for hit in hits]))
        return list(self._members_from_hits(hits))

    def eval_calltips(self):
        self.log_start()
        start_scoperef = self.get_start_scoperef()
        self.info("start scope is %r", start_scoperef)
        hits = self._hits_from_citdl(self.expr, start_scoperef)
        return self._calltips_from_hits(hits)

    def eval_defns(self):
        self.log_start()
        start_scoperef = self.get_start_scoperef()
        self.info("start scope is %r", start_scoperef)
        hits = self._hits_from_citdl(self.expr, start_scoperef, defn_only=True)
        return [ self._defn_from_hit(x) for x in hits ]

    def parent_scoperef_from_scoperef(self, scoperef,
                                      started_in_builtin_window_scope=False):
        """
        For JavaScript-in-the-browser the top-level scope is the
        Window object instance. For now we are always presuming we
        are running in the browser.

        Problem: if we *started* on the Window class then the parent
        scope should be -> built-in-blob. This is what
        'started_in_builtin_window_scope' is used for.
        """
        blob, lpath = scoperef
        if not started_in_builtin_window_scope \
           and lpath == ["Window"] and blob is self.built_in_blob:
            return None
        elif lpath:
            return (blob, lpath[:-1])
        elif blob is self.built_in_blob:
            if started_in_builtin_window_scope:
                return None
            else:
                return (self.built_in_blob, ["Window"])
        else:
            return (self.built_in_blob, [])

    _langintel = None
    @property
    def langintel(self):
        if self._langintel is None:
            self._langintel = self.mgr.langintel_from_lang(self.trg.lang)
        return self._langintel

    _libs = None
    @property
    def libs(self):
        if self._libs is None:
            self._libs = self.langintel.libs_from_buf(self.buf)
        return self._libs

    _built_in_blob = None
    @property
    def built_in_blob(self):
        if self._built_in_blob is None:
            # JS stdlib is always the last one.
            self._built_in_blob = self.libs[-1].get_blob("*")
        return self._built_in_blob

    def _hit_from_first_token(self, token, scoperef):
        """Find the token at the given or a parent scope.

        Returns the found elem and the scope at which it was found. If
        not found, this returns (None, None).
        """
        self.log("find '%s' starting at %s", token, scoperef)

        # Because we fake JavaScript classes and put the ctor
        # function inside the class, we need to push start scopes at
        # the class to the ctor. See test
        # javascript/cpln/ctor_scope_cheat for an example of why.
        elem = self._elem_from_scoperef(scoperef)
        if elem.get("ilk") == "class":
            class_name = elem.get("name")
            try:
                ctor = elem.names[class_name]
            except KeyError:
                pass
            else:
                if "__ctor__" in ctor.get("attributes", ""):
                    scoperef = (scoperef[0], scoperef[1]+[class_name])
                    self.log("push scope to class ctor %s", scoperef)

        started_in_builtin_window_scope = (scoperef[0] is self.built_in_blob
            and scoperef[1] and scoperef[1][0] == "Window")
        while 1:
            try:
                elem = self._elem_from_scoperef(scoperef)
            except KeyError, ex:
                raise EvalError("could not resolve scoperef %r: %s"
                                % (scoperef, ex))
            try:
                candidate = elem.names[token]
                if "__ctor__" in candidate.get("attributes", ""):
                    # In JavaScript we include the constructor
                    # function for a (faked) class as a method.
                    # We must skip it here or resolution of 'this'
                    # in a JS class methods will always hit the ctor
                    # instead of the class (which is by far the
                    # common case).
                    raise KeyError("skipping JavaScript ctor")
                self.log("is '%s' accessible on %s? yes", token, scoperef)
                return candidate, scoperef
            except KeyError:
                self.log("is '%s' accessible on %s? no", token, scoperef)
                scoperef = self.parent_scoperef_from_scoperef(scoperef,
                                    started_in_builtin_window_scope)
                if not scoperef:
                    return None, None

    def _members_from_hits(self, hits):
        members = set()
        for elem, scope in hits:
            # In JavaScript we include the constructor function for a
            # (faked) class as a method. Completion on an instance of
            # this class shouldn't see the ctor.
            skip_js_ctor = (elem.tag == "scope" and elem.get("ilk") == "class")

            for child in elem:
                # Only add locals when the current scope is the same
                # as the variable scope.
                attributes = child.get("attributes", "").split()
                if "__local__" in attributes:
                    # XXX: Move start_scoperef to be a part of the class
                    #start_scoperef = self.get_start_scoperef()
                    #scope_elem = start_scoperef[0]
                    #for lname in start_scoperef[1]:
                    #    if elem == scope_elem:
                    #        members.add( ("variable", child.get("name")) )
                    #        break
                    #    scope_elem = scope_elem.names[lname]
                    #else: # Don't show this variable
                    continue

                if child.tag == "scope":
                    if skip_js_ctor and child.get("ilk") == "function" \
                       and "__ctor__" in attributes:
                        continue
                    members.add( (child.get("ilk"), child.get("name")) )
                elif child.tag == "variable":
                    members.add( ("variable", child.get("name")) )
                else:
                    raise NotImplementedError("unknown hit child tag '%s': %r"
                                              % (child.tag, child))
            for classref in elem.get("classrefs", "").split():
                subhits = self._hits_from_type_inference(classref, scope)
                members.update(self._members_from_hits(subhits))
        return members

    def _calltip_from_func(self, elem):
        # See "Determining a Function CallTip" in the spec for a
        # discussion of this algorithm.
        from codeintel2.util import LINE_LIMIT
        signature = elem.get("signature")
        doc = elem.get("doc")
        ctlines = []
        if not signature:
            name = elem.get("name")
            #XXX Note difference for Tcl in _getSymbolCallTips.
            ctlines = [name + "(...)"]
        else:
            ctlines = signature.splitlines(0)
        if doc:
            ctlines += doc.splitlines(0)[:LINE_LIMIT-len(ctlines)]
        return '\n'.join(ctlines)

    def _calltip_from_class(self, elem):
        # If the class has a defined signature then use that.
        name = elem.get("name")
        signature = elem.get("signature")
        doc = elem.get("doc")
        if signature:
            ctlines = signature.splitlines(0)
            if doc:
                ctlines += doc.splitlines(0)[:LINE_LIMIT-len(ctlines)]
            return '\n'.join(ctlines)
        elif name in elem.names:
            # Typically the class element has a contructor function of
            # the same name as the class.
            ctor = elem.names[name]
            self.log("ctor is %r", ctor)
            return self._calltip_from_func(ctor)
        else:
            ctlines = [name + "(...)"]
            if doc:
                ctlines += doc.splitlines(0)[:LINE_LIMIT-len(ctlines)]
            return '\n'.join(ctlines)

    def _calltips_from_hits(self, hits):
        """
        c.f. CitadelEvaluator._getSymbolCallTips()
        """
        calltips = []

        for elem, scoperef in hits:
            #self.log("calltip for hit: %r", hit)
            if elem.tag == "variable":
                XXX
            elif elem.tag == "scope":
                ilk = elem.get("ilk")
                if ilk == "function":
                    calltips.append(self._calltip_from_func(elem))
                elif ilk == "class":
                    calltips.append(self._calltip_from_class(elem))
                else:
                    raise NotImplementedError("unexpected scope ilk for "
                                              "calltip hit: %r" % elem)
            else:
                raise NotImplementedError("unexpected elem for calltip "
                                          "hit: %r" % elem)

            ## Bug 59438: adding "(from $lpath in $file)" when helpful
            ## in calltips.
            ## TODO: Don't include all (or part) when not useful:
            ##       document.getElementsByClassName -> "(from document in
            ##       prototype)". The "document in" in not necessary.
            ## TODO: Bad with empty lpath: "(from  in prototype)"
            ## TODO: Problematic for test suite with "rand??" module names.
            ## TODO: Don't add for a local hit.
            #blobname = scoperef[0].get("name")
            #if blobname == "*":
            #    blobname = "stdlib"
            #scopename = '.'.join(scoperef[1])
            #calltips[-1] += "\n(from %s in %s)" % (scopename, blobname)
        return calltips

    def _hits_from_citdl(self, expr, scoperef, defn_only=False):
        self._check_infinite_recursion(expr)

        tokens = tuple(self._tokenize_citdl_expr(expr))
        #self.log("expr tokens: %r", tokens)

        # First part...
        hits, nconsumed = self._hits_from_first_part(tokens, scoperef)
        if not hits:
            raise CodeIntelError("could not resolve first part of '%s'" % expr)
        self.debug("_hits_from_citdl: first part: %r -> %r",
                   tokens[:nconsumed], hits)

        # ...the remainder.
        remaining_tokens = tokens[nconsumed:]
        for token in tokens[nconsumed:]:
            new_hits = []
            for elem, scoperef in hits:
                self.debug("_hits_from_citdl: resolve %r on %r in %r",
                           token, elem, scoperef)
                if token == "()":
                    try:
                        new_hits += self._hits_from_call(elem, scoperef)
                    except CodeIntelError, ex:
                        self.warn("could resolve call on %r: %s", elem, ex)
                else:
                    try:
                        new_hit = self._hit_from_getattr(
                                    elem, scoperef, token)
                    except CodeIntelError, ex:
                        self.warn(str(ex))
                    else:
                        new_hits.append(new_hit)
            hits = new_hits 

        # Resolve any variable type inferences.
        #XXX Don't we have to *recursively* resolve hits?
        #    If that is done, then need to watch out for infinite loop
        #    because _hits_from_variable_type_inference() for a variable
        #    with children just returns itself. I.e. you *can't* resolve
        #    the <variable> away.
        resolved_hits = []
        for elem, scoperef in hits:
            if elem.tag == "variable" and not defn_only:
                try:
                    subhits = self._hits_from_variable_type_inference(
                                elem, scoperef)
                except CodeIntelError, ex:
                    self.warn("could not resolve %r: %s", elem, ex)
                else:
                    resolved_hits += subhits
            else:
                resolved_hits.append( (elem, scoperef) )

        return resolved_hits

    def _hits_from_call(self, elem, scoperef):
        """Resolve the function call inference for 'elem' at 'scoperef'."""
        citdl = elem.get("returns")
        if not citdl:
            raise CodeIntelError("no return type info for %r" % elem)
        self.log("_hits_from_call: resolve '%s' for %r, scoperef: %r",
                 citdl, elem, scoperef)
        # scoperef has to be set to the function called
        func_scoperef = (scoperef[0], scoperef[1]+[elem.get("name")])
        return self._hits_from_citdl(citdl, func_scoperef)

    def _hit_from_getattr(self, elem, scoperef, token):
        """Resolve the getattr of 'token' on the given 'elem'.

        Raises CodeIntelError if could not resolve it.

        Algorithm:
        - Try to resolve it.
        - Call a hook to make an educated guess. Some attribute names
          are strong signals as to the object type -- typically those
          for common built-in classes.
        """
        self.log("resolve getattr '%s' on %r in %r:", token, elem, scoperef)
        if elem.tag == "variable":
            hits = self._hits_from_variable_type_inference(elem, scoperef)
        else:
            assert elem.tag == "scope", "elem tag is not 'scope': %r" % elem.tag
            hits = [(elem, scoperef)]

        for hit_elem, hit_scoperef in hits:
            ilk = hit_elem.get("ilk")
            if hit_elem.tag == "variable":
                attr = hit_elem.names.get(token)
                if attr is not None:
                    self.log("attr is %r on %r", attr, hit_elem)
                    var_scoperef = (hit_scoperef[0],
                                    hit_scoperef[1]+[hit_elem.get("name")])
                    return (attr, var_scoperef)
            elif ilk == "function":
                # Internal function arguments and variables should
                # *not* resolve. And we don't support function
                # attributes.
                continue
            elif ilk == "class":
                attr = hit_elem.names.get(token)
                if attr is not None:
                    self.log("attr is %r on %r", attr, hit_elem)
                    class_scoperef = (hit_scoperef[0],
                                      hit_scoperef[1]+[hit_elem.get("name")])
                    return (attr, class_scoperef)
                for classref in hit_elem.get("classrefs", "").split():
                    base_hits = self._hits_from_type_inference(classref,
                                                               hit_scoperef)
                    for base_elem, base_scoperef in base_hits:
                        if token in base_elem.names:
                            self.log("is '%s' from %s base class? yes", token,
                                     base_elem)
                            new_scoperef = (base_scoperef[0], base_scoperef[1]+
                                            [base_elem.get("name")])
                            return (base_elem.names[token], new_scoperef)
                        self.log("is '%s' from %s base class? no", token,
                                 base_elem)
            else:
                raise NotImplementedError("unexpected scope ilk: %r" % ilk)
        raise CodeIntelError("could not resolve '%s' getattr on %r in %r"
                             % (token, elem, scoperef))

    ###################
    # XPCOM example:
    # Components.classes["@mozilla.org/observer-service;1"].(getService)(Components.interfaces.nsIObserverService)
    _re_for_xpcom_citdl = re.compile(r"""Components\.classes\[["'](.*)["']"""
                                     "\]\.(getService|createInstance)\("
                                     "Components\.interfaces\.(.*)\)")
    def _hits_from_xpcom(self, citdl):
        """Parse up the xpcom expression to determine cid and iid.

        Where cid is the contract identifier
        Where iid is the interface identifier
        """

        self.log("resolve xpcom '%s'", citdl)
        match = self._re_for_xpcom_citdl.match(citdl)
        if match:
            elem = None
            groups = match.groups()
            if groups and len(groups) == 3:
                iid = groups[2]
                cid = groups[0]
                if groups[1] == "getService":
                    elem = create_xpcom_wrapper(iid, cid, "getService")
                elif groups[1] == "createInstance":
                    elem = create_xpcom_wrapper(iid, cid, "createInstance")
                if elem is not None:
                    return [(elem, [])]
        return []
    ###################

    def _hits_from_variable_type_inference(self, elem, scoperef):
        """Resolve the type inference for 'elem' at 'scoperef'."""
        assert elem.tag == "variable"
        if len(elem) != 0:
            # This is CIX for a JavaScript custom Object instance: a
            # common pattern in JS. See test javascript/cpln/local2.
            return [(elem, scoperef)]
        citdl = elem.get("citdl")
        if not citdl:
            raise CodeIntelError("no type-inference info for %r" % elem)
        ######################
        # Devel work
        if citdl.startswith("Components.classes["):
            return self._hits_from_xpcom(citdl)
        ######################
        self.log("resolve '%s' type inference for %r:", citdl, elem)
        return self._hits_from_citdl(citdl, scoperef)

    def _hits_from_type_inference(self, citdl, scoperef):
        """Resolve the 'citdl' type inference at 'scoperef'."""
        self.log("resolve '%s' type inference:", citdl)
        return self._hits_from_citdl(citdl, scoperef)

    def _hits_from_first_part(self, tokens, scoperef):
        """Resolve the first part of the expression.
        
        If the first token is found at the global or built-in level (or
        not found at all locally) then it may be a shared namespace with
        other files in the execution set. Get that down to a list of
        hits and a remaining list of expression tokens.
        """
        elem, scoperef = self._hit_from_first_token(tokens[0], scoperef)

        if (elem is None            # first token wasn't found
            or not scoperef[1]      # first token was found at global level
            # first token was found on built-in Window class (the top scope)
            or (scoperef[1] == ['Window'] and scoperef[0].get("name") == "*")
           ):
            # Search symbol table in execution set.
            #
            # Example: 'myPet.name.toLowerCase()' and 'myPet' is found
            # at top-level. First lookup 'myPet.name.toLowerCase'
            # (everything up to first '()'), in execution set, then
            # 'myPet.name', then 'myPet'. The last one should always hit
            # in current file, at least.
            for first_call_idx, token in enumerate(tokens):
                if token == "()":
                    break
            else:
                first_call_idx = len(tokens)

            hits = []
            for nconsumed in range(first_call_idx, 0, -1):
                lpath = tokens[:nconsumed]
                for lib in self.libs:
                    self.log("lookup '%s' in %s", '.'.join(lpath), lib)
                    hits_here = lib.hits_from_lpath(lpath, self.ctlr,
                                                    curr_buf=self.buf)
                    if hits_here:
                        hits += hits_here
                if hits:
                    break
            if elem is not None:
                if not hits or nconsumed == 1:
                    hits.insert(0, (elem, scoperef))
                    nconsumed = 1
                else:
                    # Hits were found in the libs that are deeper than
                    # the hit in the local buf: we need to adjust the
                    # local hit.
                    new_elem = elem
                    for token in tokens[1:nconsumed]:
                        try:
                            new_elem = new_elem.names[token]
                        except KeyError:
                            break
                    else:
                        if new_elem not in (e for e,sr in hits):
                            new_scoperef = (scoperef[0], tokens[:nconsumed-1])
                            hits.insert(0, (new_elem, new_scoperef))
        else:
            hits = [(elem, scoperef)]
            nconsumed = 1

        return hits, nconsumed


#####################################
# Devel work
#####################################

# Dictionary of already created xpcom wrappers, used so we do not keep
# creating xpcom instances / services
_g_xpcom_wrappers = {}
try:
    from xpcom import components
    _xpcom_ = True
except ImportError:
    _xpcom_ = False
from ciElementTree import Element, SubElement

def create_xpcom_wrapper(iid, cid, instanceType):
    if not _xpcom_:
        return None

    elem = _g_xpcom_wrappers.get(iid)
    if elem is not None:
        return elem

    try:
        iid = getattr(components.interfaces, iid)
        if instanceType == "getService":
            o = components.classes[cid].getService(iid)
        else:
            o = components.classes[cid].createInstance(iid)
    except:
        # Could not create one if these instances
        return None

    elem = Element("scope", name=iid, ilk="class")
    for name in o._name_to_interface_name_:
        try:
            x = getattr(o, name)
        except:
            continue  # Try the next one
        if isinstance(x, types.MethodType):
            SubElement(elem, "scope", name=name, ilk="function")
        else:
            var = SubElement(elem, "variable", name=name)
            if isinstance(x, (str, unicode)):
                var.attrib["citdl"] = "String"
            elif isinstance(x, types.BooleanType):
                var.attrib["citdl"] = "Boolean"
            elif isinstance(x, int):
                var.attrib["citdl"] = "Number"

    # Cache it
    _g_xpcom_wrappers[iid] = elem

    return elem
